package cc.fascinated.bat.features.birthday.profile;

import cc.fascinated.bat.common.Serializable;
import cc.fascinated.bat.features.birthday.UserBirthday;
import cc.fascinated.bat.model.BatGuild;
import com.google.gson.Gson;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.Setter;
import net.dv8tion.jda.api.entities.Guild;
import net.dv8tion.jda.api.entities.Member;
import net.dv8tion.jda.api.entities.User;
import net.dv8tion.jda.api.entities.channel.concrete.TextChannel;
import org.bson.Document;

import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

/**
 * @author Fascinated (fascinated7)
 */
@Getter
@Setter
@NoArgsConstructor
public class BirthdayProfile extends Serializable {
    private static final String DEFAULT_MESSAGE = "Happy Birthday {user} :tada: :birthday: You are now {age} years old!";

    /**
     * The list of birthdays that are being tracked
     */
    private Map<String, UserBirthday> birthdays;

    /**
     * The channel ID of the birthday feed
     */
    private String channelId;

    /**
     * The message to send when it is a user's birthday
     */
    private String message = DEFAULT_MESSAGE;

    /**
     * Adds a birthday to be tracked
     *
     * @param userId   the id of the user to track
     * @param birthday the birthday of the user
     */
    public void addBirthday(String userId, UserBirthday birthday) {
        if (birthdays == null) {
            birthdays = new HashMap<>();
        }
        birthdays.put(userId, birthday);
    }

    /**
     * Removes a birthday from being tracked
     *
     * @param userId the id of the user to remove
     */
    public void removeBirthday(String userId) {
        if (birthdays == null) {
            birthdays = new HashMap<>();
        }
        birthdays.remove(userId);
    }

    /**
     * Gets the birthday of a user
     *
     * @param userId the id of the user
     * @return the birthday of the user
     */
    public UserBirthday getBirthday(String userId) {
        if (birthdays == null) {
            birthdays = new HashMap<>();
        }
        return birthdays.get(userId);
    }

    /**
     * Checks if the profile has a channel setup
     *
     * @return if the profile has a channel setup
     */
    public boolean hasChannelSetup() {
        return channelId != null;
    }

    /**
     * Validates the profiles configuration
     *
     * @param guild the guild to validate
     * @return if the profile is valid
     */
    public boolean validate(BatGuild guild) {
        if (birthdays == null) {
            birthdays = new HashMap<>();
        }
        if (channelId == null) {
            return false;
        }
        if (guild.getDiscordGuild().getTextChannelById(channelId) == null) {
            channelId = null;
            return false;
        }
        return true;
    }

    /**
     * Checks to see if it is any user's birthday
     *
     * @param guild the guild to check for
     */
    public void checkBirthdays(BatGuild guild) {
        if (!validate(guild)) {
            return;
        }

        // Get today's day and month
        Calendar today = Calendar.getInstance();
        int todayDay = today.get(Calendar.DAY_OF_MONTH);
        int todayMonth = today.get(Calendar.MONTH); // Note: January is 0

        for (Map.Entry<String, UserBirthday> entry : birthdays.entrySet()) {
            Date birthday = entry.getValue().getBirthday();

            if (birthday == null) {
                continue;
            }

            Calendar birthdayCalendar = Calendar.getInstance();
            birthdayCalendar.setTime(birthday);

            int birthdayDay = birthdayCalendar.get(Calendar.DAY_OF_MONTH);
            int birthdayMonth = birthdayCalendar.get(Calendar.MONTH);

            // Check if the birthday matches today's day and month
            if (birthdayDay == todayDay && birthdayMonth == todayMonth) {
                notifyBirthday(guild, entry.getKey());
            }
        }
    }

    /**
     * Notifies a user that it is their birthday
     *
     * @param guild  the guild
     * @param userId the id of the user
     */
    private void notifyBirthday(BatGuild guild, String userId) {
        Guild discordGuild = guild.getDiscordGuild();
        Member member = discordGuild.getMemberById(userId);
        if (member == null) {
            return;
        }

        TextChannel channel = discordGuild.getTextChannelById(channelId);
        if (channel == null) { // this should never happen
            channelId = null;
            return;
        }
        channel.sendMessage(getBirthdayMessage(member.getUser())).queue();
    }

    /**
     * Gets the birthday message for a user
     *
     * @param user the user to get the message for
     * @return the birthday message
     */
    public String getBirthdayMessage(User user) {
        return message
                .replace("{user}", user.getAsMention())
                .replace("{age}", String.valueOf(birthdays.get(user.getId()).calculateAge()));
    }

    @Override
    public void reset() {
        birthdays.clear();
        channelId = null;
    }

    @Override
    public void load(Document document, Gson gson) {
        birthdays = new HashMap<>();
        for (String key : document.keySet()) {
            UserBirthday userBirthday = new UserBirthday();
            userBirthday.load((Document) document.get(key), gson);
            birthdays.put(key, userBirthday);
        }
        channelId = document.getString("channelId");
        message = (String) document.getOrDefault("message", DEFAULT_MESSAGE);
    }

    @Override
    public Document serialize(Gson gson) {
        Document document = new Document();
        for (String key : birthdays.keySet()) {
            document.put(key, birthdays.get(key).serialize(gson));
        }
        document.put("channelId", channelId);
        document.put("message", message);
        return document;
    }
}
